//----------------------------------------------------------------------------
// Copyright (C) 2003  Rafael H. Bordini, Jomi F. Hubner, et al.
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
// 
// To contact the authors:
// http://www.inf.ufrgs.br/~bordini
// http://www.das.ufsc.br/~jomi
//----------------------------------------------------------------------------


package custom.stdlib;

import custom.common.ConfidenceException;
import custom.common.Match;
import custom.common.Util;
import jason.asSemantics.DefaultInternalAction;
import jason.asSemantics.TransitionSystem;
import jason.asSemantics.Unifier;
import jason.asSyntax.Atom;
import jason.asSyntax.InternalActionLiteral;
import jason.asSyntax.ListTerm;
import jason.asSyntax.ListTermImpl;
import jason.asSyntax.Literal;
import jason.asSyntax.LiteralImpl;
import jason.asSyntax.LogExpr;
import jason.asSyntax.LogicalFormula;
import jason.asSyntax.NumberTerm;
import jason.asSyntax.Plan;
import jason.asSyntax.PlanBody;
import jason.asSyntax.PlanBodyImpl;
import jason.asSyntax.Pred;
import jason.asSyntax.StringTermImpl;
import jason.asSyntax.Structure;
import jason.asSyntax.Term;
import jason.asSyntax.Trigger;
import jason.asSyntax.VarTerm;
import jason.asSyntax.PlanBody.BodyType;

/**
  <p>Internal action:
  <b><code>custom.stdlib.adapt_plans(<i>Trigger</i>, <i>Label</i>)</code></b>.
  </p>
  
  <p>Description: adapt plans</p>
  <p>Examples:<ul> 

  //### TODO add an example
  <li> <code>custom.stdlib.adapt_plans(ListOfPlans, TFunctor, AdaptedPlans, OLabel);</code>: .</li>

  </ul>
  </p>
  
 */
public class adapt_plans extends DefaultInternalAction {
    
    @Override public int getMinArgs() { return 7; }
    @Override public int getMaxArgs() { return 7; }

    @Override
    public Object execute(TransitionSystem ts, Unifier un, Term[] args) throws Exception {
    	
    	//### The list of relevant plans - all these plans still use their original functor, which is not recognized by
    	//### the agent who requested new plans to handle the event
    	VarTerm plans = (VarTerm) args[0];
    	
    	//### The functor to be restored
    	Atom oldFunctor = (Atom) args[1];
    	
    	//### The ontology annotation of the original trigger
    	Atom ont_label = (Atom)((VarTerm) args[3]).getValue();
    	
    	Atom alignment_method = (Atom)((VarTerm) args[4]).getValue();
    	
    	float threshold = (float) ((NumberTerm)((VarTerm) args[5]).getValue()).solve();
    	
    	//### DEBUG
    	// System.out.println("Converted threshold: " + threshold);
    	
    	//### I know the first argument is a list of plans
    	ListTerm plans_list = (ListTerm) plans.getValue();
    	
    	//### Creating a new variable to contain the list of adapted plans
    	ListTerm adapted_plans = new ListTermImpl();
    	
    	String hash = "" + ((Atom)((VarTerm)args[6]).getValue()).toString();
    	
    	// System.out.println("Agent name: " + ts.getUserAgArch().getAgName());
    	
    	String label = "cool_" + hash + "_" + ts.getUserAgArch().getAgName();

    	//### Change the functor for all plans in the list
    	for (Term plan : plans_list) {
    		Plan p = (Plan) plan;
    		Literal lab = Literal.parseLiteral(""+label);
    		// p.setLabel(new StringTermImpl(label));
    		p.setLabel(new Pred(lab));
    		try {
    			adapted_plans.add(adaptPlan(p, oldFunctor, ont_label, alignment_method, threshold));
    		} catch (ConfidenceException e) {
    			System.out.println("Couldnt adapt a plan because of low confidence in matching");
    		}
    	}
    	
      return un.unifies(args[2], adapted_plans);
    }
    
    /**
     * Adapts a plan changing its functor to $oldFunctor and possibly adapting its content
     * @param plan
     * @param oldFunctor
     * @param ont_label the label of the external ontology 
     * @param threshold 
     * @return
     */
    private Plan adaptPlan(Plan plan, Atom oldFunctor, Atom ont_label, Atom alignment_method, float threshold) throws ConfidenceException {
    	//### Extract the plan's trigger (which contains the functor which must be changed) 
    	Trigger t = plan.getTrigger();
		
    	//### Start building the new trigger, creating a new predicate with the old functor and adding terms and annotations
		Pred newTriggerLiteral = new Pred(oldFunctor);
		newTriggerLiteral.addTerms(t.getLiteral().getTerms());
		newTriggerLiteral.addAnnots(t.getAnnots());
		
		//### Create the actual trigger, using information from the old trigger for the operator and type (+!,-?,...)
		//### but with the old functor
		Trigger newTrigger = new Trigger(t.getOperator(), t.getType(), newTriggerLiteral);
		
		//### Might need to adapt the context as well, if it contains terms 
		LogicalFormula old_context = plan.getContext();
		LogicalFormula new_context;
		if (old_context != null) {
			new_context = adaptContext(old_context, ont_label, alignment_method, threshold);
			if (new_context == null) {
				//### TODO this actually never happens, since if no suitable matches are fount
				//### the old context is returned without its ontology annotations
				throw new ConfidenceException("context couldn't be matched - confidence too low");
			}
		} else {
			new_context = plan.getContext();
		}
		
		//### Extract and adapt the plan body in case it is needed
		PlanBody old_pb = plan.getBody();
		PlanBody new_pb = adaptPlanBody(old_pb, ont_label, alignment_method, threshold);
		// PlanBody new_pb = old_pb;
		
		//### Finally create the actual plan, replacing the trigger, context and body with their adapted versions 
		Plan adaptedPlan = new Plan(plan.getLabel(), newTrigger, new_context, new_pb);
		
		return adaptedPlan;
    }
    
    /**
     * Adapt a plan body, replacing all occurrences of instructions who refer to ontologies which are not known by the agent
     * who requested the plan
     * @param old
     * @param ont_label
     * @return
     */
    private PlanBody adaptPlanBody(PlanBody old, Atom ont_label, Atom alignment_method, float threshold) {
    	//### old is NEVER null/empty
    	
    	// PlanBody tmp_pb = old.clonePB();
    	PlanBody tmp_pb = old;
    	
    	//### Iterates over all terms which compose the plan body
    	while (tmp_pb != null) {
    		Term t = tmp_pb.getBodyTerm();
    		BodyType type = tmp_pb.getBodyType();
    		
    		//### Why only these types?
    		if (type.equals(BodyType.achieve) || type.equals(BodyType.achieveNF)) {
    			Pred new_t = (Pred)( adaptLiteral((Pred)t, ont_label, alignment_method, threshold));
    			tmp_pb.setBodyTerm(new_t);
    		}
    		
			tmp_pb = tmp_pb.getBodyNext();
		}
    	
    	return old;
    }
    
    /**
     * Converts literals in context looking for matching concepts (if needed)
     * @param old_context
     * @param ont_label
     * @return
     */
    private LogicalFormula adaptContext(LogicalFormula old_context, Atom ont_label, Atom alignment_method, float threshold) {

    	//### DEBUG
    	//System.out.println("old context: " + old_context);
    	
    	if (old_context instanceof LogExpr) {
    		//### It's a logical expression
    		// System.out.println("instanceof");
			LogExpr old_c = (LogExpr) old_context;
	    	switch(old_c.getArity()) {
	    	case 1:
	    		//### It's a negated formula
	    		return new LogExpr(old_c.getOp(), adaptContext(old_c.getLHS(), ont_label, alignment_method, threshold));
	    	case 2:
	    		//### It's a binary formula (& or |) where both terms may be binary structures
	    		// System.out.println("adapting binary");
	    		return new LogExpr(adaptContext(old_c.getLHS(), ont_label, alignment_method, threshold), old_c.getOp(), adaptContext(old_c.getRHS(), ont_label, alignment_method, threshold));
	    	}
		} else {
			//### It's a single literal
			LiteralImpl old_c = (LiteralImpl) old_context;
			return adaptLiteral(old_c, ont_label, alignment_method, threshold);
		}
    	
    	
    	return null;
    }
    
    /**
     * Returns the ontology label if found, null otherwise
     * @param p
     * @return
     */
    private Atom hasOntologyAnnotation(Pred p) {
    	
    	//### Has no annotation, returning null 
    	if (!p.hasAnnot()) return null;
    	
    	for (Term ann : p.getAnnots()) {
    		if (ann.isStructure()) {
    			Structure ann_s = (Structure) ann;
    			if (ann_s.getFunctor().equals("o")) return (Atom)ann_s.getTerm(0);
    		}
    	}
    	return null;
    }
    
    /**
     * Adapt a literal, changing its functor
     * TODO should change also arguments
     * @param l
     * @param ont_label -> the label of the ontology known by the requesting agent 
     * @return
     */
    private Literal adaptLiteral(Literal l, Atom ont_label, Atom alignment_method, float threshold) {
    	//### This variable will contain the ontology label of the literal which is being adapted 
    	Atom old_ont;
    	
    	System.out.println("Adapting literal: " + l);
    	
    	//### IF l is a predicate AND has an ontology annotation AND this annotation is not the same known by the
    	//### requesting agent
    	if (l.isPred() && (old_ont = hasOntologyAnnotation((Pred)l)) != null && !old_ont.equals(ont_label)) {
    		
    		//### Cast the literal to Pred
    		Pred old_pred = (Pred) l;
    		
    		//### Retrieve the old functor
    		String old_functor = old_pred.getFunctor();
    		// System.out.println("old func: " + old_functor);
    		
    		//### Look for a matching concept
    		Match m = Util.match_concept(old_ont.toString(), old_functor, ont_label.toString(), alignment_method.toString(), threshold);
    		if (m.matchFound()) {

    			
    			//### Create a new predicate whose functor is the matched concept
    			Pred new_pred = new Pred(m.getMatching_concept());
    			
    			//### TODO should check recursively all terms and adapt them
    			//### add old terms to the new predicate
    			new_pred.addTerms(old_pred.getTerms());
    			
    			//### Creates the new annotation structure
    			Structure new_ont_annot = new Structure("o");
    			new_ont_annot.addTerm(ont_label);
    			
    			//### Adds the annotation to the predicate
    			new_pred.addAnnot(new_ont_annot);
    			
    			return new_pred;
    		} else {
    			// System.out.println("match NOT found");
    			//### If no match is found, the original literal is returned
    			//### removing ontology (ALL) annotations
    			//### TODO should remove only ontology annotation
    			l.clearAnnots();
    			return l;
    		}
    	} else {
    		// System.out.println("No need for a match");
    		//### The literal either is not a predicate, has no ontology annotation or has the same ontology annotation as the requested triggering event  
    		return l;	
    	}
    	
    }
    
    
    
}
