package server;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.LinkedList;
import java.util.List;
import java.util.Properties;
import java.util.Set;

import org.semanticweb.owl.align.AlignmentException;
import org.semanticweb.owl.align.Cell;

import server.alignment.Aligner;
import server.alignment.AromaAligner;
import server.alignment.JWNLAligner;


import fr.inrialpes.exmo.align.impl.URIAlignment;
import fr.inrialpes.exmo.align.parser.AlignmentParser;

class Match {
	private String concept;
	
	private double strength;
	
	public Match(String concept, double strength) {
		super();
		this.concept = concept;
		this.strength = strength;
	}

	public String getConcept() {
		return concept;
	}

	public void setConcept(String concept) {
		this.concept = concept;
	}

	public double getStrength() {
		return strength;
	}

	public void setStrength(double strength) {
		this.strength = strength;
	}

	
	
	
}

public class OntologyServerWorker {

	private OntologyServer OServer;

	public OntologyServerWorker() {

	}

	private OntologyServer getOntologyServer() {
		return OServer;
	}

	public Result handleRequest(String request, OntologyServer OServer)
			throws URISyntaxException {
		this.OServer = OServer;

		System.out.println("Request: " + request);

		String[] params = request.split(" ");
		String action = params[0];
		Result res;

		// ### Actions
		if (action.equals("REGISTER_AGENT")) {
			res = registerAgentAction(params);
			return res;
		}

		if (action.equals("ALIGN_ONTOLOGIES")) {
			res = alignOntologiesAction(params);
			return res;
		}

		if (action.equals("ECHO")) {
			res = echoAction(params);
			return res;
		}

		if (action.equals("REGISTER_ONTOLOGY")) {
			res = registerOntologyAction(params);
			return res;
		}

		if (action.equals("MATCH_CONCEPT")) {
			res = matchConceptAction(params);
			return res;
		}

		// ### Must throw exception
		return null;
	}

	/**
	 * Aligns the two ontologies previously registered with labels o1_label and
	 * o2_label The alignment is a RDF file stored locally
	 * 
	 * @param o1_label
	 * @param o2_label
	 * @return
	 */
	private Result alignOntologies(String o1_label, String o2_label,
			String method) {
		URI o1_uri = getOntologyServer().getOntologyURI(o1_label);
		URI o2_uri = getOntologyServer().getOntologyURI(o2_label);
		
		String alignment_folder_path;

		Aligner a = null;

		//### Load and read configuration file, which includes the path where alignments will be saved
		Properties prop = new Properties();
	    String fileName = "oa.config";
	    InputStream is;
		
	    try {
			is = new FileInputStream(fileName);
		    prop.load(is);

		    alignment_folder_path = prop.getProperty("oa.alignment_folder_path");
			System.out.println("Ontology Artifact configuration file read, alignment folder path = " + alignment_folder_path);

			// ### choose the alignment method
			if (method.equals("jwnl")) {
				a = new JWNLAligner();
			} else if (method.equals("aroma")) {
				a = new AromaAligner();
			} else {

				// TODO notify "unknown alignment method" error
			}
			String filename = a.align(o1_label, o2_label, o1_uri, o2_uri, alignment_folder_path);

			getOntologyServer().addAlignment(o1_label, o2_label, filename);

			Result res = new Result();
			res.setAnswer("success");
			return res;
			
		} catch (FileNotFoundException e) {
			System.out.println("Ontology Artifact configuration file was not found");
			// TODO Auto-generated catch block
			e.printStackTrace();
			return null;
		} catch (IOException e) {
			System.out.println("Ontology Artifact configuration file could not be read properly");
			// TODO Auto-generated catch block
			e.printStackTrace();
			return null;
		}

		
	}

	/**
	 * Wrapper for the actual method
	 * 
	 * @param params
	 * @return
	 */
	private Result alignOntologiesAction(String[] params) {
		String o1_label = params[1];
		String o2_label = params[2];
		String alignment_method = params[3];

		// ### Aligns ontologies both ways
		alignOntologies(o2_label, o1_label, alignment_method);
		return alignOntologies(o1_label, o2_label, alignment_method);
	}

	private Result echoAction(String[] params) {
		Result res = new Result();
		res.setAnswer(params[1].toUpperCase());
		return res;
	}

	/**
	 * Register an agent
	 * 
	 * @param params
	 * @return
	 */
	private Result registerAgentAction(String[] params) {
		// ### TODO stub
		return null;
	}

	/**
	 * Register an ontology in the server
	 * 
	 * @param params
	 * @return
	 * @throws URISyntaxException
	 */
	private Result registerOntologyAction(String[] params)
			throws URISyntaxException {
		// ### TODO stub
		String ontology_uri_string = params[1];
		URI ontology_uri = new URI(ontology_uri_string);
		String label = params[2];
		System.out.println("OServer: registering ontology "
				+ ontology_uri_string + " with label " + label);

		getOntologyServer().registerOntology(label, ontology_uri);

		Result res = new Result();
		res.setAnswer("success");
		return res;
	}

	private Result matchConceptAction(String[] params) {
		String o1_label = params[1];
		String c1 = params[2];
		String o2_label = params[3];
		String alignment_method = params[4];
		float threshold = Float.parseFloat(params[5]);

		String matching_concept = concept_match(o1_label, c1, o2_label,
				alignment_method, threshold);

		Result res = new Result();
		res.setAnswer(matching_concept);
		return res;
	}

	/**
	 * Translates an atom For example, the atom "beverage" found in ontology
	 * "http://www.owl-ontologies.com/Supermarket.owl" would be translated in
	 * the full URI
	 * "class http://www.owl-ontologies.com/Supermarket.owl#Beverage" TODO
	 * 
	 * @param atom
	 * @return
	 */
	public static String atom2concept(String atom) {
		return atom.substring(0, 1).toUpperCase() + atom.substring(1);
	}

	/**
	 * the other way around
	 * 
	 * @param concept
	 * @return
	 */
	public static String concept2atom(String concept) {
		// ### TODO stub
		return concept.substring(0, 1).toLowerCase() + concept.substring(1);
	}

	/**
	 * Returns a string for a concept matching c1
	 * 
	 * @param o1_label
	 * @param c1
	 * @param o2_label
	 * @param alignment_method
	 * @return
	 */
	/*
	private String concept_match2(String o1_label, String c1, String o2_label,
			String alignment_method, float threshold) {
		String atom_as_concept = atom2concept(c1);
		String alignment_file = getOntologyServer().getAlignmentFile(o1_label,
				o2_label);

		// System.out.println("1: " + o1_label + " ,2: " + o2_label);

		// ### If there is no alignment, align the two ontologies right now and
		// proceed
		if (alignment_file == null) {
			// ### Aligns ontologies both ways
			alignOntologies(o2_label, o1_label, alignment_method);
			alignOntologies(o1_label, o2_label, alignment_method);
			alignment_file = getOntologyServer().getAlignmentFile(o1_label,
					o2_label);
		}

		System.out.println("Alignment file: " + alignment_file);

		try {
			AlignmentParser ap;
			// URI o1_uri = getOntologyServer().getOntologyURI(o1_label);

			ap = new AlignmentParser(0);
			URIAlignment al;

			al = (URIAlignment) ap.parse(alignment_file);
			URI URI1 = al.getOntology1URI();
			URI URI2 = al.getOntology2URI();

			// object in the alignment
			URI full_uri = new URI(URI1 + "#" + atom_as_concept);

			// System.out.println("aac" + atom_as_concept);
			
			//### Create a structure to contain cells

			Cell cell = al.getAlignCell1(full_uri);

			// ### Trying without uppercasing the first character
			if (cell == null) {
				full_uri = new URI(URI1 + "#" + c1);
				cell = al.getAlignCell1(full_uri);
			}

			// ### TODO check confidence threshold
			if (cell != null) {
				String c2 = ((URI) (cell.getObject2())).toString();
				String matching_concept = concept2atom(c2.substring(URI2
						.toString().length() + 1));
				if (cell.getStrength() >= threshold) {
					System.out.println("Found matching concept "
							+ matching_concept + " with Similarity "
							+ cell.getStrength());
					return matching_concept + " " + cell.getStrength();
				} else {
					System.out.println("Found matching concept "
							+ matching_concept + " but the confidence level "
							+ cell.getStrength()
							+ " is lower than the threshold " + threshold);
					return null;
				}

			}

		} catch (AlignmentException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (URISyntaxException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		System.out.println("No matching concept found...");
		return null;
	}
	*/
	
	/**
	 * Adds a new Match to the list mantaining descending order
	 * @param matches
	 * @param concept
	 * @param strength
	 */
	private void addValidMatch(List<Match> matches, String concept, double strength) {
		for (int i = 0; i < matches.size(); i++) {
			if (matches.get(i).getStrength() <= strength) {
				matches.add(i, new Match(concept, strength));
				return;
			}
		}
		matches.add(new Match(concept, strength));
	}
	
	/**
	 * Returns a string for a concept matching c1
	 * 
	 * @param o1_label
	 * @param c1
	 * @param o2_label
	 * @param alignment_method //### TODO is it necessary?
	 * @return
	 */
	private String concept_match(String o1_label, String c1, String o2_label,
			String alignment_method, float threshold) {
		String atom_as_concept = atom2concept(c1);
		String alignment_file = getOntologyServer().getAlignmentFile(o1_label,
				o2_label);

		// System.out.println("1: " + o1_label + " ,2: " + o2_label);

		// ### If there is no alignment, align the two ontologies right now and
		// proceed
		if (alignment_file == null) {
			// ### Aligns ontologies both ways
			alignOntologies(o2_label, o1_label, alignment_method);
			alignOntologies(o1_label, o2_label, alignment_method);
			alignment_file = getOntologyServer().getAlignmentFile(o1_label,
					o2_label);
		}

		System.out.println("Alignment file: " + alignment_file);

		try {
			AlignmentParser ap;
			// URI o1_uri = getOntologyServer().getOntologyURI(o1_label);

			ap = new AlignmentParser(0);
			URIAlignment al;

			al = (URIAlignment) ap.parse(alignment_file);
			URI URI1 = al.getOntology1URI();
			URI URI2 = al.getOntology2URI();

			// object in the alignment
			URI full_uri = new URI(URI1 + "#" + atom_as_concept);

			// System.out.println("aac" + atom_as_concept);
			
			//### Create a structure to contain cells indicating a strong enough match
			List<Match> valid_cells = new LinkedList<Match>();

			// Cell cell = al.getAlignCell1(full_uri);
			
			Set<Cell> candidate_cells = al.getAlignCells1(full_uri);
			
			/*
			System.out.println("full uri" + full_uri);
			System.out.println("cell" + cell );
			System.out.println("candidate cells" + candidate_cells );
			*/
			
			// ### Trying without uppercasing the first character (for properties)
			if (candidate_cells == null) {
				full_uri = new URI(URI1 + "#" + c1);
				candidate_cells = al.getAlignCells1(full_uri);
			}
			
			if (candidate_cells != null) {
				for (Cell c : candidate_cells) {
					String c2 = ((URI) (c.getObject2())).toString();
					String matching_concept = concept2atom(c2.substring(URI2
							.toString().length() + 1));
					if (c.getStrength() >= threshold) {
						System.out.println("Found matching concept "
								+ matching_concept + " with Similarity "
								+ c.getStrength());
						addValidMatch(valid_cells, matching_concept, c.getStrength());
						// return matching_concept + " " + c.getStrength();
					} else {
						System.out.println("Found matching concept "
								+ matching_concept + " but the confidence level "
								+ c.getStrength()
								+ " is lower than the threshold " + threshold);
					}
				}
					
			}
			
			//### If any valid match is found return the result
			if (valid_cells.size() > 0) {
				String res = "";
				for (Match m : valid_cells) {
					res += m.getConcept() + " " + m.getStrength() + " ";
					res += m.getConcept() + " " + (m.getStrength()-0.1) + " ";
				}
				return res.trim();
			}

		} catch (AlignmentException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (URISyntaxException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		
		System.out.println("No matching concept found...");
		return null;
	}
	

	/*
	 * private String concept_match_bak(String o1_label, String c1, String
	 * o2_label) { String atom_as_concept = atom2concept(c1); String
	 * alignment_file = getOntologyServer().getAlignmentFile(o1_label,
	 * o2_label);
	 * 
	 * System.out.println("1: " + o1_label + " ,2: " + o2_label);
	 * 
	 * //### If there is no alignment, align the two ontologies right now and
	 * proceed if (alignment_file == null) { alignOntologies(o1_label,
	 * o2_label); alignment_file =
	 * getOntologyServer().getAlignmentFile(o1_label, o2_label); }
	 * 
	 * System.out.println("Alignment file: " + alignment_file);
	 * 
	 * try { AlignmentParser ap; URI o1_uri =
	 * getOntologyServer().getOntologyURI(o1_label);
	 * 
	 * ap = new AlignmentParser(0); URIAlignment al;
	 * 
	 * al = (URIAlignment) ap.parse(alignment_file); URI URI1 =
	 * al.getOntology1URI(); URI URI2 = al.getOntology2URI();
	 * 
	 * if (al.getFile1().equals(o1_uri)) { // ### Ontology O1 is the FIRST //
	 * object in the alignment URI full_uri = new URI(URI1 + "#" +
	 * atom_as_concept);
	 * 
	 * Cell cell = al.getAlignCell1(full_uri);
	 * 
	 * // ### TODO check confidence threshold if (cell != null) { //
	 * System.out.println("STR: " + cell.getStrength()); String c2 = ((URI)
	 * (cell.getObject2())).toString(); String matching_concept =
	 * concept2atom(c2.substring(URI2 .toString().length() + 1));
	 * System.out.println("Found matching concept " + matching_concept +
	 * " with Similarity " + cell.getStrength()); return matching_concept + " "
	 * + cell.getStrength(); } } else { // ### Ontology O1 is the SECOND object
	 * in the alignment URI full_uri = new URI(URI2 + "#" + atom_as_concept);
	 * 
	 * Cell cell = al.getAlignCell2(full_uri);
	 * 
	 * // ### TODO check confidence threshold if (cell != null) {
	 * 
	 * String c2 = ((URI) (cell.getObject1())).toString(); String
	 * matching_concept = concept2atom(c2.substring(URI1 .toString().length() +
	 * 1)); System.out.println("Found matching concept " + matching_concept +
	 * " with Similarity " + cell.getStrength()); return matching_concept + " "
	 * + cell.getStrength(); } } } catch (AlignmentException e) { // TODO
	 * Auto-generated catch block e.printStackTrace(); } catch
	 * (URISyntaxException e) { // TODO Auto-generated catch block
	 * e.printStackTrace(); }
	 * 
	 * System.out.println("No matching concept found..."); return null; }
	 */
}
